#include "zlib_src\\zlib.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define USE_CONSOLE
#include "allegro.h"

// disable MS VC6 warning "unreferenced inline function has been removed"
#pragma warning(disable : 4514)

// new handy types
typedef signed char        BYTE;
typedef unsigned char      UBYTE;
typedef short int          WORD;
typedef unsigned short int UWORD;
typedef long               DWORD;
typedef unsigned long      UDWORD;

// new structs, with NO padding between members
#pragma pack(1)

   typedef struct PAK_HEADER_S
   {
      UBYTE signature[3];
      UBYTE version;
   } PAK_HEADER_S;

   typedef struct PAK_TEX_HEADER_S
   {
      UDWORD nb_files;
      UDWORD u04;
      UBYTE  padding[244];
   } PAK_TEX_HEADER_S;

   typedef struct PAK_TEX_FILEINFO_HEADER_S
   {
      UDWORD type_id;
      UDWORD offset;
      UDWORD compressed_size;
   } PAK_TEX_FILEINFO_HEADER_S;

   typedef struct PAK_TEX_IMAGEINFO_HEADER_S
   {
      char   filename[32];
      UWORD  width;
      UWORD  height;
      UBYTE  repeat_type_id;
      UDWORD repeat_compr_size;
      UBYTE  padding[39];
   } PAK_TEX_IMAGEINFO_HEADER_S;

#pragma pack()


// global datas
int  glb_no_convert = FALSE;
int  glb_cmap[16];


// ================================================================================
// ZLIB 16bpp
// ================================================================================
int extract_tga_type4(FILE * in, long length, int width, int height, char * filename)
{
   BITMAP * bmp;
   UBYTE  * src_buf, * dst_buf;
   long   size, cur_pos, out_len;
   int    status, r, g, b, a, c;
   UWORD  x, y, * word_ptr, word;
   FILE   * out;


   // create src buffer
   src_buf = (UBYTE *) malloc(length);
   if (src_buf == NULL)
   {
      printf("can't allocate %i bytes for SRC buffer of %s\n", length, filename);
      return -1;
   }

   // fill src buffer
   cur_pos = ftell(in);
   fread(src_buf, length, 1, in);
   fseek(in, cur_pos, SEEK_SET);

   if (glb_no_convert == TRUE)
   {
      out = fopen(filename, "wb");
      if (out == NULL)
      {
         printf("ERROR, can't create %s for output\n", filename);
         return -1;
      }
      fwrite(src_buf, 1, length, out);
      fclose(out);
      return 0;
   }

   // create dst buffer
   size = width * height * 2;
   size *= 3; // just for safety
   dst_buf = (UBYTE *) malloc(size);
   if (dst_buf == NULL)
   {
      free(src_buf);
      printf("can't allocate %i bytes for DST buffer of %s\n", size, filename);
      return -1;
   }

   // try the decompression
   out_len = size;
   status = uncompress(dst_buf, (UDWORD *) (& out_len), src_buf, length);

   // if decompression was a success, create image file
   if (status == Z_OK)
   {
      if (out_len != (width * height * 2))
      {
         printf("ERROR : out_len != width * height * 2\n");
         printf("        %lu     != %i * %i * 2 (%li)\n",
            out_len, width, height, width * height * 2);
      }

      bmp = create_bitmap_ex(32, width, height);
      if (bmp == NULL)
      {
         printf("can't create bitmap of %i * %i pixels 32bpp\n", width, height);
         free(dst_buf);
         free(src_buf);
         return -1;
      }

      word_ptr = (UWORD *) dst_buf;
      for (y=0; y < height; y++)
      {
         for (x=0; x < width; x++)
         {
            word = * word_ptr;

            a = (word & 0xF000) >> 12;
            r = (word & 0x0F00) >> 8;
            g = (word & 0x00F0) >> 4;
            b = word & 0x000F;

            a = glb_cmap[a];
            r = glb_cmap[r];
            g = glb_cmap[g];
            b = glb_cmap[b];

            c = makeacol_depth(32, r, g, b, a);
            putpixel(bmp, x, y, c);

            // next pixel
            word_ptr++;
         }
      }
      save_tga(filename, bmp, NULL);
      destroy_bitmap(bmp);
   }

   // end
   free(dst_buf);
   free(src_buf);
   return 0;
}


// ================================================================================
// raw 32bpp
// ================================================================================
int extract_tga_type6(FILE * in, long length, int width, int height, char * filename)
{
   BITMAP  * bmp;
   long    cur_pos;
   int     r, g, b, a, c;
   int     x, y;
   UDWORD  * udword_ptr, udword;
   UBYTE   * src_buf;
   FILE    * out;


   // create src buffer
   src_buf = (UBYTE *) malloc(length);
   if (src_buf == NULL)
   {
      printf("can't allocate %i bytes for SRC buffer of %s\n", length, filename);
      return -1;
   }

   // fill src buffer
   cur_pos = ftell(in);
   fread(src_buf, length, 1, in);
   fseek(in, cur_pos, SEEK_SET);

   if (glb_no_convert == TRUE)
   {
      out = fopen(filename, "wb");
      if (out == NULL)
      {
         printf("ERROR, can't create %s for output\n", filename);
         return -1;
      }
      fwrite(src_buf, 1, length, out);
      fclose(out);
      return 0;
   }

   // final image
   bmp = create_bitmap_ex(32, width, height);
   if (bmp == NULL)
   {
      free(src_buf);
      printf("can't create bitmap of %i * %i pixels 32bpp\n", width, height);
      return -1;
   }

   udword_ptr = (UDWORD *) src_buf;

   for (y=0; y < height; y++)
   {
      for (x=0; x < width; x++)
      {
         udword = * udword_ptr;

         a = (udword & 0xFF000000LU) >> 24;
         r = (udword & 0x00FF0000LU) >> 16;
         g = (udword & 0x0000FF00LU) >> 8;
         b = udword  & 0x000000FFLU;

         c = makeacol_depth(32, r, g, b, a);
         putpixel(bmp, x, y, c);

         // next pixel
         udword_ptr++;
      }
   }
   save_tga(filename, bmp, NULL);
   destroy_bitmap(bmp);

   // end
   free(src_buf);
   return 0;
}


// ================================================================================
void list_tex(FILE * pak)
{
   PAK_TEX_HEADER_S           tex_header;
   PAK_TEX_FILEINFO_HEADER_S  fileinfo_header;
   PAK_TEX_IMAGEINFO_HEADER_S imageinfo_header;
   UDWORD                     i, * offset, * compsize, size, comp_size, off_size, * type_id;
   unsigned long              w, h;
   int                        c, padding_ok;
   long                       progress_width = 67, progress_max, progress_curr,
                              progress_next;
   char                       new_filename[40];


   if (pak == NULL)
      return;

   printf("decoding as TEX\n");

   // tex header
   fread( & tex_header, sizeof(PAK_TEX_HEADER_S), 1, pak);
   printf("nb files : 0x%08lX = %lu\n", tex_header.nb_files, tex_header.nb_files);
   printf("unknown  : 0x%08lX = %lu\n", tex_header.u04, tex_header.u04);
   printf("(skip 244 bytes of padding)\n");
   padding_ok = TRUE;
   for (c=0; c < 244; c++)
   {
      if (tex_header.padding[c] != 0)
      {
         padding_ok = FALSE;
         break;
      }
   }
   if (padding_ok == FALSE)
   {
      printf("ERROR : 244 bytes aren't all for padding");
      for (c=0; c < 244; c++)
      {
         if ((c % 16) == 0)
            printf("\n%02X :", c);
         else if ((c % 8) == 0)
            printf(" -", c);
         printf(" %02X", tex_header.padding[c]);
      }
      printf("\n");
   }

   // type_ids list
   size = tex_header.nb_files * sizeof(UDWORD);
   type_id = (UDWORD *) malloc(size);
   if (type_id == NULL)
   {
      printf("can't allocate %i bytes for %lu type_ids\n", size, tex_header.nb_files);
      return;
   }
   memset(type_id, 0, size);

   // offsets list
   size = tex_header.nb_files * sizeof(UDWORD *);
   offset = (UDWORD *) malloc(size);
   if (offset == NULL)
   {
      printf("can't allocate %i bytes for %lu offsets\n", size, tex_header.nb_files);
      return;
   }
   memset(offset, 0, size);

   // copressed sizes list
   size = tex_header.nb_files * sizeof(UDWORD);
   compsize = (UDWORD *) malloc(size);
   if (compsize == NULL)
   {
      printf("can't allocate %i bytes for %lu compressed sizes\n", size, tex_header.nb_files);
      return;
   }
   memset(compsize, 0, size);

   // file infos
   for (i=0; i < tex_header.nb_files; i++)
   {
      fread( & fileinfo_header, sizeof(PAK_TEX_FILEINFO_HEADER_S), 1, pak);
      printf("\nfile %u\n", i);
      printf("   type id         : %lu\n", fileinfo_header.type_id);
      printf("   offset          : 0x%08lX = %lu\n", fileinfo_header.offset, fileinfo_header.offset);
      printf("   compressed size : 0x%08lX = %lu\n", fileinfo_header.compressed_size, fileinfo_header.compressed_size);

      fflush(stdout);
      type_id[i]  = fileinfo_header.type_id;
      offset[i]   = fileinfo_header.offset;
      compsize[i] = fileinfo_header.compressed_size;
   }

   // image infos
   progress_max  = tex_header.nb_files; 
   progress_curr = 0;
   fprintf(stderr, "progress : ");
   fflush(stderr);
   for (i=0; i < tex_header.nb_files; i++)
   {
      fseek(pak, offset[i], SEEK_SET);
      printf("\nfile %u @ offset 0x%08lX (%li)\n", i, ftell(pak), ftell(pak));

      comp_size = compsize[i];
      off_size  = 0;
      if (i < (tex_header.nb_files - 1))
         off_size = offset[i+1] - ftell(pak);

      printf("file datas size... compressed = 0x%08lX (%li), in_file = 0x%08lX (%li)\n",
         comp_size, comp_size, off_size, off_size);

      fread( & imageinfo_header, sizeof(PAK_TEX_IMAGEINFO_HEADER_S), 1, pak);
      w = imageinfo_header.width;
      h = imageinfo_header.height;
      printf("   filename : %.32s\n", imageinfo_header.filename);
      printf("   width    : %lu\n",    w);
      printf("   height   : %lu\n",    h);
      printf("   rep_t_id : 0x%02X (%u)\n", imageinfo_header.repeat_type_id, imageinfo_header.repeat_type_id);
      printf("   rep_c_sz : 0x%08lX (%li)\n", imageinfo_header.repeat_compr_size, imageinfo_header.repeat_compr_size);
      printf("   (skip 39 bytes of padding)\n");
      padding_ok = TRUE;
      for (c=0; c < 39; c++)
      {
         if (imageinfo_header.padding[c] != 0)
         {
            padding_ok = FALSE;
            break;
         }
      }
      if (padding_ok == FALSE)
      {
         printf("ERROR : 39 bytes aren't all for padding");
         for (c=0; c < 39; c++)
         {
            if ((c % 16) == 0)
               printf("\n%02X :", c);
            else if ((c % 8) == 0)
               printf(" -", c);
            printf(" %02X", imageinfo_header.padding[c]);
         }
         printf("\n");
      }
      if (comp_size != imageinfo_header.repeat_compr_size)
         printf("ERROR : 2 compressed sizes not the same\n");
      if (type_id[i] != imageinfo_header.repeat_type_id)
         printf("ERROR : 2 type_ids not the same\n");

      fflush(stdout);

      // extract files
      if (i >= 1)
      {
         //sprintf(new_filename, "%05lu-%s", i, imageinfo_header.filename);
         strcpy(new_filename, imageinfo_header.filename);

         // extract the image
         if (imageinfo_header.repeat_type_id == 6)
         {
            // raw 32bpp
            if (extract_tga_type6(pak, comp_size, w, h, new_filename) == 0)
               printf("extraction of %s ok\n", new_filename);
            else
               printf("ERROR : extraction of %s was impossible\n", new_filename);
         }
         else if (imageinfo_header.repeat_type_id == 4)
         {
            // zlib 16bpp
            if (extract_tga_type4(pak, comp_size, w, h, new_filename) == 0)
               printf("extraction of %s ok\n", new_filename);
            else
               printf("ERROR : extraction of %s was impossible\n", new_filename);
         }
         else
         {
            printf("ERROR : unknown type\n");
         }
      }

      // progress
      progress_next = i * progress_width / progress_max ;
      if (progress_next != progress_curr)
      {
         for (;progress_curr < progress_next; progress_curr++)
         {
            fprintf(stderr, ".");
            fflush(stderr);
         }
      }
   }
   fprintf(stderr, "ok\n");
   fflush(stderr);

   free(compsize);
   free(offset);
   free(type_id);
}


// ================================================================================
int main(int argc, char ** argv)
{
   PAK_HEADER_S pak_header;
   FILE         * pak;
   int          x, y;


   printf("sacred_tex_extractor v 1.00 (2 May 2006) freeware, by Paul Siramy\n");
   printf("=================================================================\n\n");

   // helper to transform 4bpp into 8bpp color componant values
   for (x=0; x < 16; x++)
   {
      y = 255 * x;
      glb_cmap[x] = y / 15;
      if ((y % 15) >= 8)
         glb_cmap[x]++;
   }

   if (argc < 2)
   {
      printf("syntaxe : sacred_tex_extractor.exe \"<file.pak>\" [-no-convert]\n");
      return -1;
   }
   if (argc == 3)
   {
      if (stricmp(argv[2], "-no-convert") == 0)
         glb_no_convert = TRUE;
   }

   // open file
   pak = fopen(argv[1], "rb");
   if (pak == NULL)
   {
      printf("can't open %s\n", argv[1]);
      return -1;
   }

   // read pak header
   fread( & pak_header, sizeof(PAK_HEADER_S), 1, pak);

   printf("file      : %s\n\n", argv[1]);
   printf("signature : %.3s\n", pak_header.signature);
   printf("version   : %i\n\n", pak_header.version);

   if ((strnicmp((char *) pak_header.signature, "TEX", 3) == 0) &&
       (pak_header.version == 3))
   {
      // PAK\texture*.pak
      allegro_init();
      list_tex(pak);
   }
   else
   {
      printf("don't know how to decode (signature and/or version not suported\n");
   }

   // end
   fclose(pak);
   return 0;
}
